package extension

import (
	"context"

	"git.perx.ru/perxis/perxis-go"
	"git.perx.ru/perxis/perxis-go/pkg/collections"
	"git.perx.ru/perxis/perxis-go/pkg/content"
	"git.perx.ru/perxis/perxis-go/pkg/errors"
	"git.perx.ru/perxis/perxis-go/pkg/schema"
	"git.perx.ru/perxis/perxis-go/pkg/setup"
	pb "git.perx.ru/perxis/perxis-go/proto/extensions"
)

const (
	StateNotInstalled = pb.State_NOT_INSTALLED
	StatePending      = pb.State_PENDING
	StateInstalled    = pb.State_INSTALLED
	StateInProgress   = pb.State_IN_PROGRESS
	StateFail         = pb.State_FAIL

	MetadataKey = "extension"
)

type (
	InstallRequest   = pb.InstallRequest
	CheckRequest     = pb.CheckRequest
	UninstallRequest = pb.UninstallRequest

	ExtensionDescriptor = pb.ExtensionDescriptor
	State               = pb.State
)

var (
	ErrStart = errors.New("start failed")
	ErrStop  = errors.New("stop failed")

	ErrInstall   = errors.New("install failed")
	ErrUpdate    = errors.New("update failed")
	ErrCheck     = errors.New("check failed")
	ErrUninstall = errors.New("uninstall failed")

	ErrNotInstalled     = errors.New("not installed")
	ErrUnknownExtension = errors.New("unknown extension")

	ManifestAssets   = perxis.NewAssets[*ExtensionDescriptor]()
	ManifestFromFile = ManifestAssets.MustOneFrom
)

// Runnable описывает интерфейс сервиса с запуском и остановкой. Вызывается сервером расширений
type Runnable interface {
	Start() error
	Stop() error
}

// Extension описывает интерфейс расширения Perxis
type Extension interface {
	// GetDescriptor возвращает описание расширения
	GetDescriptor() *ExtensionDescriptor

	// Install вызывается при установке расширения в пространство
	Install(ctx context.Context, in *InstallRequest) error

	// Check вызывается для проверки состояния расширения
	Check(ctx context.Context, in *CheckRequest) error

	// Uninstall вызывается для удаления расширения из пространства
	Uninstall(ctx context.Context, in *UninstallRequest) error

	// Action вызывается для выполнения расширением действия
	Action(ctx context.Context, in *ActionRequest) (*ActionResponse, error)
}

func CheckInstalled(ctx context.Context, content *content.Content, spaceID, envID, extension string) (bool, error) {
	status, err := NewStorage(content, nil).GetExtension(ctx, spaceID, envID, extension)
	if err != nil {
		return false, err
	}
	return status.State == StateInstalled, nil
}

func isMetadataEqual(s1, s2 *schema.Schema) bool {
	if s1.Metadata == nil && s2.Metadata == nil {
		return true
	}
	if s1.Metadata == nil || s2.Metadata == nil {
		return false
	}
	return s1.Metadata[MetadataKey] == s2.Metadata[MetadataKey]
}

// UpdateCollectionStrategy В дополнение к стратегии по умолчанию делает проверку, что обновляемая
// коллекция была установлена расширением. Если в метаданных схемы отсутствует специальный ключ `MetadataKey`,
// это означает, что коллекция была создана пользователем и отношения к расширению не имеет - вернется ошибка.
func UpdateCollectionStrategy(s *setup.Setup, exist, collection *collections.Collection) (*collections.Collection, bool, bool, error) {
	if !s.IsForce() && !collection.IsView() && !exist.IsView() && !isMetadataEqual(collection.Schema, exist.Schema) {
		return nil, false, false, errors.WithDetailf(collections.ErrAlreadyExists, "Коллекция с идентификатором '%s' "+
			"уже существует. Удалите ее или вызовите установку расширения с флагом Force", collection.ID)
	}
	return setup.DefaultUpdateCollectionStrategyFn(s, exist, collection)
}
