package test

import (
	"os"
	"testing"

	"git.perx.ru/perxis/perxis-go/pkg/extension"
	"git.perx.ru/perxis/perxis-go/pkg/references"
	"git.perx.ru/perxis/perxis-go/pkg/schema"
	"git.perx.ru/perxis/perxis-go/pkg/schema/field"
	"git.perx.ru/perxis/perxis-go/pkg/schema/modify"
	"git.perx.ru/perxis/perxis-go/pkg/schema/validate"
	"github.com/stretchr/testify/require"
)

func TestFromFiles(t *testing.T) {
	t.Run("Non-existen path", func(t *testing.T) {
		schemas, err := schema.FromFS(os.DirFS("non-existen"))
		require.Error(t, err)
		require.ErrorContains(t, err, "no such file or directory")
		require.Nil(t, schemas)
	})

	t.Run("Success", func(t *testing.T) {
		schemas, err := schema.FromFS(os.DirFS("assets"))
		for _, s := range schemas {
			s.ClearState()
		}
		require.NoError(t, err)
		require.Len(t, schemas, 2, "В директории хранятся две корректные схемы")
		require.ElementsMatch(t, []*schema.Schema{getPagesSchema(), getPagesSchema()}, schemas, "Cхемы должны соответствовать объекту из функции")
	})
}

// Оригинальное объявление схемы Web/Страницы
// Значение констант подставлено вручную
func getPagesSchema() *schema.Schema {
	content := field.Object(
		true, "blocks", field.Array(
			references.Field([]string{"web_block_*"}).SetTitle("Блок"),
		).
			SetTitle("Блоки").
			WithUI(&field.UI{
				Widget: "BlockList",
				Options: map[string]interface{}{
					"create": map[string]interface{}{
						"classes": []string{"class_web_blocks"},
					},
				},
			}),
	).SetTitle("Содержимое страницы")

	// SEO
	seo := field.Object().WithIncludes("web_seo").SetTitle("SEO")

	//Settings
	settings := field.Object(true,
		"name", field.String(
			validate.Required(),
			modify.Value("$perxis.Item.Template == true ? _value : replace_markers(_value)"),
		).SetTitle("Название").SetTextSearch(true).SetIndexed(true),
		"parent", field.String().SetTitle("Родительский раздел").SetDescription("Раздел сайта, где расположена страница").WithUI(&field.UI{
			Widget: "Lookup",
			Options: map[string]interface{}{
				"allowedCollections": []interface{}{
					map[string]interface{}{"collection": "web_pages"},
				},
			},
		}), // TODO lookup "section == true"
		"path", field.String(
			modify.Value("$perxis.Item.Template == true ? _value : parent == nil ?  '/' : slug == nil ? make_path(parent, slugify(replace_markers(name))) : make_path(parent, slugify(replace_markers(slug)))"),
		).SetTitle("Путь").SetUnique(true).SetTextSearch(true), // TODO readonly
		"slug", field.String(modify.Value("$perxis.Item.Template == true ? _value : parent == nil ? nil : _value == nil ?  slugify(replace_markers(name)) :  slugify(replace_markers(_value))")).SetTitle("Slug").SetDescription("Идентификатор страницы в адресе URL").SetTextSearch(true),
		"section", field.Bool().SetTitle("Раздел").WithUI(&field.UI{Widget: "Checkbox"}),
		"datasource", field.String().WithUI(&field.UI{
			Widget: "Lookup",
			Options: map[string]interface{}{
				"allowedCollections": []interface{}{
					map[string]interface{}{"collection": "web_datasources"},
				},
			},
		}).SetTitle("Источник данных").SetDescription("Источник данных из которого будут формироваться подстраницы раздела").SetCondition("section == true"),
		"redirect", field.Bool().SetTitle("Перенаправление").SetDescription("Страница не имеет содержимого и перенаправляет посетителей на другой адрес").WithUI(&field.UI{Widget: "Checkbox"}),
		"redirect_path", field.String().SetTitle("Страница для перехода").WithUI(&field.UI{
			Widget: "Lookup",
			Options: map[string]interface{}{
				"allowedCollections": []interface{}{
					map[string]interface{}{"collection": "web_pages"},
				},
			},
		}).SetCondition("redirect == true"),
		"redirect_url", field.String().SetTitle("URL для перехода").SetCondition("redirect == true"),
		"outputs", field.Array(field.String().WithUI(&field.UI{
			Widget: "Lookup",
			Options: map[string]interface{}{
				"allowedCollections": []interface{}{
					map[string]interface{}{"collection": "web_outputs"},
				},
			},
		})).SetTitle("Форматы страницы для вывода"),
		"navTitle", field.String().SetTitle("Название для навигации"),
		"navHide", field.Bool().SetTitle("Скрыть страницу из навигации").WithUI(&field.UI{Widget: "Checkbox"}),
		"weight", field.Number(field.NumberFormatInt).SetTitle("Порядок следования").WithUI(&field.UI{Widget: "NumberInput"}),
	).SetTitle("Настройки")

	// Advanced
	advanced := field.Object(
		"scripts", field.Array(
			field.Object(
				"src", field.String().SetTitle("URL для загрузки скрипта"),
				"type", field.String().SetTitle("Media Type скрипта"),
				"content", field.String().SetTitle("Содержимое скрипта"),
			),
		).SetTitle("Дополнительные скрипты"),
	).SetTitle("Расширенные настройки")

	// Design
	design := field.Object().WithIncludes("web_design").SetTitle("Дизайн")

	//Variables
	variables := field.Object(true, "variables", field.Array(getVarsField())).SetTitle("Переменные")

	// Page
	page := schema.New(
		"content", content,
		"seo", seo,
		"settings", settings,
		"advanced", advanced,
		"design", design,
		"variables", variables,
	).WithMetadata(extension.MetadataKey, "perxisweb")

	// Includes
	page.SetIncludes(
		field.Include{Ref: "hoop_item_options", Optional: true},
		field.Include{Ref: "ext_web_pages_*", Optional: true},
	)

	//UI
	page.Field.UI.ListView = &field.View{Options: map[string]interface{}{
		"fields":    []string{"path", "name", "updated_at", "updated_by", "state"},
		"sort":      []string{"path"},
		"page_size": 50,
	}}
	page.Field.UI.Options["title"] = "name"
	page.Field.UI.Options["key"] = "path"
	page.Field.UI.Options["description"] = "path"
	page.Field.UI.Widget = "Tabs"
	page.Field.UI.Options["collection_icon"] = "ApartmentOutlined/FileTextOutlined"

	_ = page.ConvertTypes()
	page.ClearState()
	return page
}

func getVarsSchema() (sch *schema.Schema) {
	sch = schema.New(
		"id", field.String(validate.Required()).SetTextSearch(true).SetTitle("Идентификатор переменной"),
		"name", field.String().SetTextSearch(true).SetTitle("Название переменной"),
		"value", field.String().SetTitle("Значение переменной"),
	).WithMetadata(extension.MetadataKey, "perxisweb")
	//UI
	sch.Field.UI.ListView = &field.View{Options: map[string]interface{}{
		"fields":    []string{"id", "name", "updated_at", "updated_by", "state"},
		"page_size": 50,
	}}

	sch.Field.UI.Options["collection_icon"] = "SettingOutlined/FileExcelOutlined"
	return sch
}

func getVarsField() *field.Field {
	return &getVarsSchema().Field
}
